package GreenCompiler;

import java.util.*;
import java.io.*;
import AST.*;
import CompilerError.*;
import Interface.*;


/**
  In directory basiclib should be put arrays of all basic types up to dimension
  three ( array(byte)[], array(byte)[][], array(byte)[][][], ... ), the basic 
  classes (char, byte, integer, ...), all classes used by the 
  array classes (Iter$Array$1$byte, ...).
      In directory exceptions should be put all exception classes. 
      In directory greenlib should be  put the basic green library (Vector, Filter, ...)
      
  Grammar for the project
    Project ::= FileNameList  *OPTIONS OptionsList 
    FileNameList ::= | FileNameList FileName
    OptionsList ::= Option | OptionsList Option
    Option ::= 
       BASICLIB "=" FileName | 
       LIB "=" FileName | 
       WRAPPER "=" FileName | 
       ARRAYDIR "=" FileName | 
       CLASSDIR "=" FileName | 
       JAVALIB "=" FileName | 
       TEMPLATEDIR  "=" FileName | 
       PROJECTDIR "=" FileName | 
       SUBTYPEEQUALSUBCLASS | 
       INTERFACEDIR "=" FileName | 
       USEIRL | 
       COMPILEONLY |
       COMPILEALL
    FileName ::= STRING | PATHNAME
    
*/


public class GreenProject {
    
    public boolean init( 
            String fileName, 
            ProjectError error,  
            // ErrorMsg errorMsg,
            ErrorMsgList errorMsg,
            ClassErrorListList errorListList 
            ) {
        this.errorMsg = errorMsg;
        this.error = error;
        this.errorListList = errorListList;
        MyFile myFile = new MyFile(fileName);
        input = myFile.readFile();
        if ( myFile.getError() != MyFile.ok_e ) {
           error.show("Cannot open file " + fileName);
           return false;
        }
        setDefaultOptions();
        arrayPathLib = new Vector();
        arrayJavaLib = new Vector();
        dirList = new Vector();
        //c classList = new ClassList();
        toBeCompiledList = new Vector();
        toCreateClassList = new ToCreateClassList();
        toCreateArrayList = new ToCreateArrayList();
        separator = File.separator;
        projectDir = "";
        return true;
    }
    
    private void setDefaultOptions() {
        subtypeEqualSubclass = false;
        arrayDir = "arraytmp/";
        classDir = "classtmp/";
        interfaceDir = "interface/";
        templateDir = null;
        compilerOptions = new CompilerOptions();
        compilerOptions.setUseIRL(false);
        compilerOptions.setSubtypeEqualSubclass(false);
        compilerOptions.setCompileOnly(false);
        compilerOptions.setCompileAll(false);
    }

 
    private boolean compileClass( ClassFile cf, ClassTypeInterface cti ) {
       /* compile the class cf whose interface is cti. Generates code for
          the class and class object and all interfaces */
       
       ClassErrorList classErrorList = null;
       String fileName = cf.getClassName() + ".g";
       
       /*if ( cf.getClassName().compareTo("Program") == 0 ) 
          System.out.println("Program");  */
       
       

       String pathFile = cf.getDirectory() + separator + fileName;
       MyFile myFile = new MyFile(pathFile);
       input = myFile.readFile();
       if ( myFile.getError() != MyFile.ok_e )  {
          error.show("Can't read file " + pathFile);
          return false;
       }
       else {
          MyLexer lexer;
          
          String className;
             // builds the name of the file to be generated
          if ( NameServer.basicTypes.get(cti.getName()) != null ||
                  // if the class is a basic class or String
               cti.getName().compareTo("String") == 0 )  
             className = NameServer.buildBasicTypeClassName(cti.getName());
          else 
             className = cti.getJavaName();
            
          String outClassName = cf.getDirectory() + separator + className + ".java";
          String outClassObjectName = cf.getDirectory() + separator + 
                 cti.getClassObjectInterface().getInterfaceName() + ".java";
          File f_g = new File(cf.getDirectory() + separator + cf.getClassName() + ".g");
          File f_java = new File(outClassName);
          File f_co_java = new File(outClassObjectName);

          if ( compilerOptions.getCompileAll() ||
               ! f_java.exists() ||
               f_g.lastModified() > f_java.lastModified() ||
               cti.getIsShellClass() ||
               ! f_co_java.exists() ||
               f_g.lastModified() > f_co_java.lastModified() ) {
             try {
               lexer = new MyLexer( input, errorMsg );
               errorMsg.setLexer(lexer);
               //# errorMsg.setFileName(pathFile);
               classErrorList = new ClassErrorList(cf.getDirectory(), fileName);
               errorMsg.setClassErrorList(classErrorList);
               //#
               lexer.next();
             } catch ( Exception e ) {
               error.show("Error reading file " + pathFile);
               errorListList.add(classErrorList);
               errorMsg.removeClassErrorList();
               return false;
             }
             
             MyParser parser;
   
             try { 
                parser = new MyParser(lexer, errorMsg, cti, 
                    til, compilerOptions);
             } catch ( Throwable e ) {
               errorListList.add(classErrorList);
               errorMsg.removeClassErrorList();
               return false;
             }
   
             GenericClass aClass;
             try {
               aClass = (GenericClass ) parser.parse();
             } catch ( Exception e ) {
               System.out.println(e);
               errorListList.add(classErrorList);
               errorMsg.removeClassErrorList();
               //error.show("Error compiling " + pathFile);
               return false;
             }
            
             if ( aClass == null ) {
                errorListList.add(classErrorList);
                errorMsg.removeClassErrorList();
                return false;
             }
   
          
             /*if ( fileName.compareTo("B.g") == 0 ) {
                System.out.println("Compiling ");     
             }
             */
                        
             try {
                if ( compilerOptions.getCompileAll() ||
                     ! f_java.exists() ||
                     f_g.lastModified() > f_java.lastModified()  ) {
                   PrintWriter outClass = new PrintWriter( 
                       new FileOutputStream(outClassName) );
                    
                   //if ( basicClassesTable.get(aClass.getName()) != null ) 
                        // if not a basic class, compile it.
                      aClass.gen(outClass, compilerOptions, className);
                      //@ aClass.genInterfaces(projectDir + separator + interfaceDir, errorMsg);
                      aClass.genInterfaces(interfaceDir + separator, errorMsg);
                      //}
                      outClass.close();
                }
                if ( cti.getIsShellClass() ) {
                       String fileName_ivc = cf.getDirectory() + separator + 
                       aClass.getShell_ivcName() + ".java";
                   try {
                      PrintWriter out_ivc = new PrintWriter( 
                            new FileOutputStream(fileName_ivc) );
                      aClass.genShell_ivc(out_ivc);
                      out_ivc.close();
                   }
                   catch ( Exception excep ) {
                      errorMsg.signal("Error writing " + fileName_ivc);
                   }
                } 
                else {
                   ClassObject classObject = aClass.getClassObject();
                   if ( compilerOptions.getCompileAll() ||
                        ! f_co_java.exists() ||
                        f_g.lastModified() > f_co_java.lastModified() ) {
                      PrintWriter outClassObject = new PrintWriter( 
                         new FileOutputStream(outClassObjectName) );
                      classObject.gen(outClassObject, compilerOptions);
                      outClassObject.close();
                         // only generates interfaces if the class
                         // is generated.
                      //@ classObject.genInterfaces(projectDir + separator + interfaceDir, errorMsg);
                      classObject.genInterfaces(interfaceDir + separator, errorMsg);
                   }
                }
               
             }
             catch( Exception excep ) {
                error.show("Error writing to file " + outClassName + " or " + outClassObjectName + " or internal error");
                //excep.printStackTrace();
                errorMsg.removeClassErrorList();
                errorListList.add(classErrorList);
                return false;
             }
            errorMsg.removeClassErrorList();
            errorListList.add(classErrorList);
          }
          return true;
       }
    }



    
    public void make() {

        ClassFile cf;
        
          // number of the current line
        lineNumber = 1;
        tokenPos = 0;
        beforeOptions = true;
        inFileName = true;
        cfl = new Vector();
        ccl = new Vector();
        til = new TypeInterfaceList();
        pcl = new Hashtable();
        
        try { 
           nextToken();
           parse();
        } catch ( Exception e ) {
            return ;
        }
        Enumeration e;
        if ( error.getWasACompilerError() || errorMsg.getWasAnCompilerError() ) 
           return;
           
            
           // add library classes to cfl
        boolean result = addArrayLib();
        if ( ! addBasicLib() )
           result = false;
        if ( ! addWrapperLib() ) 
           result = false;
        
        
           /* repeat for all classes cited in the project.
              if the .gi class exists and .g does not, just read
              the .gi file. Assume the file has already been
              compiled. If the .gi file does not exist or is 
              older than the .g file or .java does not exist, 
              inserts the class into the list of classes to be 
              compiled.
                Therefore not all classes of cfl list will be 
              compiled.
           */
        e = cfl.elements();
        while ( e.hasMoreElements() ) {
           cf = (ClassFile ) e.nextElement();
           String directory = cf.getDirectory();
           String name = cf.getClassName();
           
           /* if ( name.compareTo("Program") == 0 ) 
              System.out.println("at  Program");   */
              
           File fg = new File(directory + separator + name + ".g");
           File fgi = new File(directory + separator + name + ".gi");
           File fj = new File(directory + separator + 
              NameServer.buildRealClassJavaName(name) + ".java");
           
              
           if ( ! fg.exists() ) {
                 // .g file does not exists
               if ( ! fgi.exists() ) {
                     // .gi files does not exist
                   error.show("File " + fgi.getPath() + " cannot be read", -1);
               }
               else 
                     // .gi file exists
                  try {
                     readInterface(cf, true);
                  } catch ( Exception ce ) {
                  }
           }
           else 
                 // fg exists
               if ( compilerOptions.getCompileAll() ||
                    ! fgi.exists() || ! fj.exists() || 
                    fg.lastModified() > fgi.lastModified() ||
                    fg.lastModified() > fj.lastModified() ) {
                   // Either .gi or .java does not exists or .gi is older than ".g". 
                   // Therefore, ".g" needs to be recompiled
                   try { 
                      if ( readInterface(cf, false) ) {
                         ccl.addElement(cf);
                      }
                   } catch ( Exception ce ) {
                   }
               }
               else {
                   // .gi exists and .gi is newer than .g
                   try {
                      readInterface(cf, true);
                   } catch ( Exception ce ) {
                   }
               }
        }
         
        if ( error.getWasACompilerError() || errorMsg.getWasAnCompilerError() ) 
           return;
        // create instantiations of parameterized classes and array classes
        createClassesFromClassList();
        if ( error.getWasACompilerError() || errorMsg.getWasAnCompilerError() ) 
           return;
        
        
        //#errorMsg.setShowNameLineColumn(false);
        errorMsg.removeClassErrorList();
        //#

           // reify all interfaces
        try {
           reifyBasicClasses();
        } catch ( CompilerException ce ) {
            return ;
        }
        if ( error.getWasACompilerError() || errorMsg.getWasAnCompilerError()  ) 
           return;
        try { 
           e = til.elements();
           while ( e.hasMoreElements() ) {
               String interfaceName = (String ) e.nextElement();
              ((ClassTypeInterface ) til.get(interfaceName)).reify(til, errorMsg);
           }
        } catch ( CompilerException ce ) {
           return ;
        }
        catch ( Exception exc ) {
           error.show("Internal error at GreenProject");
           return ;
        }
        if ( error.getWasACompilerError() || errorMsg.getWasAnCompilerError() ) 
           return;
           
        //# errorMsg.setShowNameLineColumn(true);
        
        //moved from after the next while
        createDirectories();
        
        e = ccl.elements();
        while ( e.hasMoreElements() ) {
           cf = (ClassFile ) e.nextElement();
//           if ( cf.getClassName().compareTo("A") == 0 ) {
           ClassTypeInterface cti = ((ClassTypeInterface ) til.get( cf.getClassName() ));
           compileClass(cf, cti);
//                }
        }
        //end moved from after the next while
        
        e = ccl.elements();
        while ( e.hasMoreElements() ) {
           cf = (ClassFile ) e.nextElement();
           ClassTypeInterface cti = ((ClassTypeInterface ) til.get( cf.getClassName() ));
           if ( cti == null ) {
              errorMsg.show("Internal error: cti == null");
           }
           if ( ! cti.getIsShellClass() ) {
              String filename = cf.getDirectory() + separator + cf.getClassName() + ".gi";

              try {
                 File file_gi = new File(filename);
                 File file_g  = new File(cf.getDirectory() + separator + cf.getClassName() + ".g");
                 if ( ! file_gi.exists() ||
                        file_g.lastModified() > file_gi.lastModified() ) {
                    PrintWriter outInterface = new PrintWriter( 
                       new FileOutputStream(filename) );
                    cti.genI(outInterface);
                    outInterface.close();
                 }
              }
              catch( Exception excep) {
                 error.show("Error writing to file " + filename);
              }
           }
        }
           
        if ( error.getWasACompilerError() || errorMsg.getWasAnCompilerError()  ) 
           return;
        prepareBatFile();
        System.out.println("This is the end !!!");
    }

   private boolean prepareBatFile() {
         // prepare the do.bat and run.bat files. "do.bat" compiles,
         // the java files generated by the Green compiler. 
         // "run.bat" executes the program

         // add all paths to dirList. They will after be include 
         // at "-sourcepath" of the Java compiler
      dirList.addElement( projectDir );
      //@ dirList.addElement( projectDir + separator + arrayDir );
      //@ dirList.addElement( projectDir + separator + classDir );
      //@ dirList.addElement( projectDir + separator + interfaceDir );
      dirList.addElement( arrayDir );
      dirList.addElement( classDir );
      dirList.addElement( interfaceDir );

      dirList.addElement( pathBasicLib );
      dirList.addElement( pathWrapper );
      Enumeration e = arrayPathLib.elements();
      while ( e.hasMoreElements() ) 
         dirList.addElement( (String ) e.nextElement() );

         // create file  "filelist.txt" with all Java files 
         // to be compiled.
      String outFileName = projectDir + separator + "do.bat";
      try {
         PrintWriter out = new PrintWriter( 
                   new FileOutputStream(outFileName) );
                   
         out.print("javac ");
         
         out.print("-sourcepath ");
         int size = dirList.size();
         e = dirList.elements();
         while ( e.hasMoreElements() )  {
            out.print( (String ) e.nextElement() );
            if ( --size > 0 ) 
               out.print(";");
         }
         /*
         String classPath = System.getProperty("java.class.path");
         if ( classPath != null && classPath.length() > 0 )
            out.print(classPath);
         */
         
         out.println(" @filelist.txt");
         out.close();    
         
         outFileName = projectDir + separator + "filelist.txt";
         
         out = new PrintWriter( new FileOutputStream(outFileName) );

         out.println( pathBasicLib + separator + "PackedException.java");
         // all Java files that need to be compiled come from ccl
         e = ccl.elements();
         while ( e.hasMoreElements() ) {
            ClassFile cf = (ClassFile ) e.nextElement();
            if ( cf.getClassName().compareTo("String") != 0 ) {
               ClassTypeInterface cti = cf.getClassTypeInterface();
               out.println( cf.getDirectory() + separator 
                  + cf.getJavaFileName() + ".java" );
               if (  cti.getIsShellClass() ) 
                  out.println( cf.getDirectory() + separator + 
                     cti.getShell_ivcName() + ".java" );
               else
                  out.println( cf.getDirectory() + separator + 
                     NameServer.buildClassObjectName(cf.getClassName()) 
                     + ".java" );
            }
         }
         
           // class object name of the class in which the execution should start.
         String startClassObjectName = NameServer.buildClassObjectName(startClass);
         outFileName = projectDir + separator + "Main.java";
         out.println(outFileName);
         out.close();
         
           // create the Main class in which the execution 
           // will start at the Java code
         out = new PrintWriter( new FileOutputStream(outFileName) );
         ws.set(out);
         ws.println("import java.lang.reflect.*;");
         ws.println("");
         ws.println("public class Main {");
         ws.add();
         ws.println("public static void main( String args[] ) {");
         ws.add();
            // access the class object of the start class
         ws.println(startClassObjectName + 
             " start = _" + startClass + ".$classObject;" );
             // if an exception is not caught, call method 
             // throw of HCatchUncheckedException
         //ws.println("_Runtime.$classObject._setCatchUnchecked$_CatchUncheckedException( _HCatchUncheckedException.$classObject._new() );");
         ws.println("_Runtime.$classObject._$addCatchObjectStack$_Catch( _HCatchUncheckedException.$classObject._new() );");         
         ws.println("initialize();");
         
         
         ClassObjectTypeInterface coti = til.get(startClass).getClassObjectInterface();

           /* build method name of run with and without parameters.
              If there is parameter, there should be just one:
              String []args */
         String runMethodName, runStringMethodName;
         ParameterList pl = new ParameterList();
         runMethodName = NameServer.buildMethodSearchName("run", pl ); 
         
         Parameter p = new Parameter("args"); 
         ArrayType arrayType = new ArrayType( Type.stringType, 1 );
         p.setType(arrayType);
         pl.add(p);
         runStringMethodName = NameServer.buildMethodSearchName("run", pl); 
         
         MethodInterface runMethod, runStringMethod;
         runMethod = coti.publicSearch(runMethodName);
         runStringMethod = coti.publicSearch(runStringMethodName);
         if ( runMethod != null && runStringMethod != null ) {
            error.show("Error: the class object in which the execution starts, " + startClass + " should contain just only method run");
            return false;
         }
         
         ws.println("try {");         
         ws.add();
         if ( runMethod != null ) {
            ws.println("start._run();");
         }
         else
            ws.println("start._run(args);");
         ws.sub();
         ws.println("}");
         ws.println("catch( Throwable pe ) {");
         ws.add();
         ws.println("try {");
         ws.add();
         ws.println("Class argTypes[] = new Class[1];");
         ws.println("argTypes[0] = (new Object()).getClass();");
         ws.println("Object e = pe;");
         ws.println("while ( e != null && e instanceof PackedException ) {");
         ws.add();
         ws.println("e = ((PackedException ) e).getException();");
         //ws.println("System.out.println(\"Exception \" + e.getClass().getName() +");
         //ws.println("     \" not caught\");");
         ws.sub();
         ws.println("}");
         ws.println("Class aClass = _Runtime.$classObject._getCatchUnchecked().getClass();");
         ws.println("Method aMethod = aClass.getDeclaredMethod(\"select\", argTypes);");
         ws.println("Object parameters[] = new Object[1];");
         ws.println("parameters[0] = e;");
         ws.println("aMethod.invoke(_Runtime.$classObject._getCatchUnchecked(), ");
         ws.println("     parameters);");
         ws.sub();
         ws.println("} catch ( Throwable ee ) {");
         ws.add();
         ws.println("System.out.println(\"Internal error at main\");");
         ws.sub();
         ws.println("}");
         ws.sub();
         ws.println("}");
         
         /*        
         ws.println("} catch ( PackedException pe ) {");
         ws.add();
         ws.println("Object e = pe;");
         ws.println("while ( e != null && e instanceof PackedException )");
         ws.add();
         ws.println("e = ((PackedException ) e).getException();");
         ws.sub();
         ws.println("System.out.println(\"Exception \" + e.getClass().getName() +");
         ws.println("   \" not caught\");");
         ws.sub();
         ws.println("}");
         ws.println("catch ( Throwable e ) {");
         ws.add();
         ws.println("System.out.println(\"Exception \" + e.getClass().getName() +");
         ws.println("   \" not caught\");");
         ws.sub();
         ws.println("}");
         */
         
         
         ws.sub();
         ws.println("}");

         ws.println("");
         ws.println("");
         ws.println("public static void initialize() {");
         ws.add();  

         if ( compilerOptions.getUseIRL() ) {
               // call the initialize method of all classes
            Vector fileList = new Vector();
            e = dirList.elements();
            while ( e.hasMoreElements() )  {
               File file = new File( (String ) e.nextElement() );
               String allFiles[] = file.list();
               for( int i = 0; i < allFiles.length; i++ ) 
                  if (  allFiles[i].endsWith(".java") ) {
                     String className = allFiles[i].substring(0, allFiles[i].lastIndexOf('.'));
                        // skips interface classes, class object classes
                     if ( ! className.startsWith("$co$") &&
                          ! className.startsWith("$mi$") &&
                          (className.charAt(0) == '_' ||
                           className.charAt(0) == '$')  ) {
                        ws.println(className + ".initialize();");
                        fileList.addElement(className);
                     } 
                  }
            }

              // put all information on all classes in an array stored
              // at runtime
            ws.println("_Runtime.$classObject._$setClasses$$Array$1$_ClassInfo(");
            ws.add();
            ws.println("new $Array$1$_ClassInfo (");
            ws.add();
            ws.println("new _ClassInfo[] {");
            ws.add();
           
            size = fileList.size();
            e = fileList.elements();
            while ( e.hasMoreElements() ) {
               String className = (String ) e.nextElement();
               if ( className.indexOf("$shell") >= 0 ) 
                 size--;
               else {
                  ws.print( className + ".$i" );
                  if ( --size > 0 )
                    out.println(",");
                  else
                     out.println();
               }
            }
            ws.println("}");
            ws.sub();
            ws.println(")");
            ws.sub();
            ws.println(");");
            ws.sub();     
         }
         
         ws.sub();
         ws.println("}");
         
         ws.sub();
         out.println("}");
         out.close();

            // created file run.bat that executes the program
         outFileName = projectDir + separator + "run.bat";
         out = new PrintWriter( 
                   new FileOutputStream(outFileName) );
         out.print("java -classpath ");
         size = dirList.size();
         e = dirList.elements();
         while ( e.hasMoreElements() )  {
            out.print( (String ) e.nextElement() );
            if ( --size > 0 ) 
               out.print(";");
         }
         out.println(" Main");
         out.close();
         
      }
      catch( Exception excep ) {
         // System.out.println("Exception " + excep.toString());
         // excep.printStackTrace();
         error.show("Error writing to file " + outFileName);
         return false;
      }
      return true;
   }
        
        
   private void createDirectories() {
    
      //@ File arrayDirFile = new File(projectDir + separator + arrayDir);
      File arrayDirFile = new File(arrayDir);
      //@ File classDirFile = new File(projectDir + separator + classDir);
      File classDirFile = new File(classDir);
      
      //@ File interfaceDirFile = new File(projectDir + separator + interfaceDir);
      File interfaceDirFile = new File(interfaceDir);
      
      if ( ! arrayDirFile.exists() )
         arrayDirFile.mkdirs();
      if ( ! classDirFile.exists() )
         classDirFile.mkdirs();
      if ( ! interfaceDirFile.exists() )
         interfaceDirFile.mkdirs();
   }


   private boolean readInterface( ClassFile cf, boolean is_gi ) {
       /*
         read the interface of a .g or .gi file. Add it to til. 
         If it is a parameterized class like List (not List(char), 
         which is an instantiated parameterized class), add it
         to pcl.
       */
    
       ClassErrorList classErrorList = null;
   
       String pathFile = cf.getDirectory() + separator + cf.getClassName();
       pathFile = pathFile + (is_gi ? ".gi" : ".g");
       
       MyFile myFile = new MyFile(pathFile);
       input = myFile.readFile();
       if ( myFile.getError() != MyFile.ok_e ) {
           error.show("Error reading file " + pathFile);
           return false;
       }
       else {
          MyLexer lexer;
          try {
            lexer = new MyLexer( input, errorMsg );
            errorMsg.setLexer(lexer);
            
            //# errorMsg.setFileName(pathFile);
            classErrorList = new ClassErrorList(cf.getDirectory(), cf.getClassName() + (is_gi ? ".gi" : ".g"));
            errorMsg.setClassErrorList(classErrorList);
            //#

            
            lexer.next();
          } catch ( Exception e ) {
            error.show("Internal error in readInterface");
            errorMsg.removeClassErrorList();
            errorListList.add(classErrorList);
            return false;
          }
          
          InterfaceParser parser;
          boolean ok;

          
          try { 
             parser = new InterfaceParser( 
                   lexer, errorMsg, toCreateClassList, 
                   toCreateArrayList, is_gi, compilerOptions ); 
                   
             ok = parser.parse();
          } catch ( Throwable e ) {
            //e.printStackTrace();
            error.show("Error in readInterface");
            errorMsg.removeClassErrorList();
            errorListList.add(classErrorList);
            return false;
          }
          if ( ! ok ) {
             errorMsg.removeClassErrorList();
             errorListList.add(classErrorList);
             return false;
          }
             
             
          ParameterizedClassInterface pci = parser.getParameterizedClassInterface();
          if ( pci != null ) {
                 // found a parameterized class
              pci.setClassFile(cf);
              if ( pcl.put( pci.getName(), pci ) != null ) {
                 error.show("Duplicate class: " + pci.getName());
              }
              errorMsg.removeClassErrorList();
              errorListList.add(classErrorList);
              return false;
          }
          else {
             ClassTypeInterface cti = parser.getInterface();
             if ( cti == null || cti.getName() == null ) {
                error.show("Internal error at readInterface");
                errorMsg.removeClassErrorList();
                errorListList.add(classErrorList);
                return false;
             }
             else { 
                cti.setClassFile(cf);
                cf.setClassTypeInterface(cti);
                /* if ( cti.getName().compareTo("Program") == 0 ) 
                   System.out.println("at A");  */
                if ( til.add( cti ) != null ) {
                   error.show("Duplicate class : " + cti.getName());
                   errorMsg.removeClassErrorList();
                   errorListList.add(classErrorList);
                   return false;
                }
             }
          }
          errorMsg.removeClassErrorList();
          errorListList.add(classErrorList);
       }
       return true;
    }

      
      
      
      
    private void createClassesFromClassList() {
        /*
          create all parameterized an array classes cited when
          reading the Green files (".g") or ".gi" files.
        */
        
        ToCreateClassList list = toCreateClassList;
        ToCreateClassList all = list;
        toCreateClassList = new ToCreateClassList();
        Hashtable allArrays = new Hashtable();
        
        while ( list.getSize() != 0 || toCreateArrayList.getSize() != 0 ) {
            Hashtable t = list.getTable();
            Enumeration e = t.keys();
            ClassType classType;
              // create parameterized classes
            while ( e.hasMoreElements() ) {
                String javaName = (String ) e.nextElement();
                classType = (ClassType ) t.get( javaName );
                createInstantiatedClass(classType);
            }
            e = toCreateArrayList.keys();
               // create array classes
            while ( e.hasMoreElements() ) {
                String arrayName = (String ) e.nextElement();
                
                if ( allArrays.get(arrayName) == null ) {
                   ArrayType anArray = toCreateArrayList.get(arrayName);
                   createArray(anArray);
                   allArrays.put(arrayName, "");
                }
            }
               // check if the new classes to be created have not been created before
            Vector removeList = new Vector();
            t = toCreateClassList.getTable();
            e = t.keys();
            while ( e.hasMoreElements() ) {
                String javaName = (String ) e.nextElement();
                classType = (ClassType ) t.get( javaName );
                if ( all.get(javaName) == null )
                   // class have not been created before
                   all.add(classType);
                else 
                   // class have been created before. Don't create it again
                   removeList.addElement(javaName);
            }
            e = removeList.elements();
            while ( e.hasMoreElements() ) 
               toCreateClassList.remove( (String ) e.nextElement());

            
            list = toCreateClassList;
            toCreateArrayList = new ToCreateArrayList();
            toCreateClassList = new ToCreateClassList();
        }
    }


    private boolean createInstantiatedClass( ClassType classType ) 
       /* create a real class from classType, which is something 
          like List(char) or Stack(Person) Create .g and .gi 
          files, put in classDir */
    {
        ParserParamClass ppc = new ParserParamClass(error, projectDir, classDir, separator);
        ClassFile outcf;
        
           // get the parameterized class, that in the .g file.
        ParameterizedClassInterface pci = (ParameterizedClassInterface ) 
               pcl.get( classType.getName() );
        if ( pci == null ) {
            error.show("Parameterized class " + classType.getName() + " not found");
            return false;
        }
          // create the real class
        if ( (outcf = ppc.create(classType, pci)) != null ) 
           try { 
                 // read the interface of the class, add it
                 // to til and ccl 
              if ( readInterface(outcf, false) ) {
                  ccl.addElement(outcf);
                  return true;
              }
           } catch ( Exception ce ) {
           }
        return false;
    }
          
    
    
    private boolean createAnyArray( ArrayType anArray ) {
        /*
          create an array "anArray" which should not be an array
          of a basic type.
        */
        
        if ( anArray.getNumDimensions() != 1 ) {
            error.show("Attempt to use an array of non-basic class with " + 
               anArray.getNumDimensions() + " dimensions" );
            return false;
        }
        String dir;
        String filename;
        
        MyFile mf;
        //@ dir = projectDir + separator + arrayDir + separator;
        dir = arrayDir + separator;
            
        if ( contentsOf_array_any_1_g == null ) {
            filename = templateDir + separator + "array-any-1.g";
            mf = new MyFile(filename);
            contentsOf_array_any_1_g = mf.readFile();
            if ( mf.getError() != MyFile.ok_e ) {
                error.show("Error reading " + mf.getName() );
                return false;
            }
            fileAny1g = new File(filename);
        }
        
        File f;
          // create an array. The name of it in Java is the same
          // as in Green.
        f = new File(dir + anArray.getJavaName() + ".g");
        if ( ! f.exists() ||
             f.lastModified() < fileAny1g.lastModified() )
           CreateArray.createArrays( contentsOf_array_any_1_g, dir,
                anArray.getJavaName(), anArray.getType().getName(), null, 
                anArray.getType().getJavaName() );

        
        //@ ClassFile outcf = new ClassFile(projectDir + separator + arrayDir, anArray.getJavaName() ); 
        ClassFile outcf = new ClassFile(arrayDir, anArray.getJavaName() ); 
        /* if ( anArray.getJavaName().compareTo("$Array$1$_ConstantInfo") == 0 ) 
           System.out.println("array"); */
        try {
           if ( readInterface(outcf, false) ) {
              ccl.addElement(outcf);
           }
        } catch ( Exception ce ) {
           return false;
        }
        

          // each array needs an iter class, created below 
        if ( contentsOf_iter_array_any_1_g == null ) {
            filename = templateDir + separator + "iter-any-1.g";
            mf = new MyFile(filename);
            contentsOf_iter_array_any_1_g = mf.readFile();
            if ( mf.getError() != MyFile.ok_e ) {
                error.show("Error reading " + mf.getName() );
                return false;
            }
            fileIterAny1g = new File(filename);
        }
          
        f = new File( dir + "AnyIter$" + anArray.getType().getJavaName() + ".g" );
        if ( ! f.exists() || 
              f.lastModified() < fileIterAny1g.lastModified() )
        
          CreateArray.createArrays( contentsOf_iter_array_any_1_g, dir,
             "AnyIter$" + anArray.getType().getJavaName(), 
             anArray.getType().getName(), null, 
             anArray.getType().getJavaName() );
             
        //@ outcf = new ClassFile(projectDir + separator + arrayDir, 
        //@   "AnyIter$" + anArray.getType().getJavaName() ); 
        outcf = new ClassFile(arrayDir, 
          "AnyIter$" + anArray.getType().getJavaName() ); 
        try { 
           if ( readInterface(outcf, false) ) {
              ccl.addElement(outcf);
           }
        } catch ( Exception ce ) {
           return false;
        }
          
        
        return true;         
    }
    
    private void createArray( ArrayType anArray ) {

        if ( ! anArray.getType().hasArray() && 
               anArray.getType().getName().compareTo("String") != 0 )
            createAnyArray(anArray);
    }
    
    
    private void createAuxArrayClasses() {
        /*
          insert classes like Iter(char), Function(char), ... into 
          the list of classes to be created.
          
        */
        
        for ( int i = 0; i < Type.arrayAnyValue.length; i++ ) {
            TypeList typeList = new TypeList();
            typeList.add( new ClassType( Type.arrayAnyValue[i].getName(), null) );
            toCreateClassList.add( new ClassType( "Iter", typeList ) );
            toCreateClassList.add( new ClassType( "Function", typeList ) );
            toCreateClassList.add( new ClassType( "Command", typeList ) );
            toCreateClassList.add( new ClassType( "Filter", typeList ) );
        }
    }
            

    private void reifyBasicClasses() {
       /*
         set the pointers of all basic classes, wrapper classes,
         Any, AnyValue, AnyClass, ... That is, set the pointers 
         of Type.*, where * represents the static private variables,
         to the correct interfaces.
         
       */
       ClassTypeInterface cti;
       for ( int i = 0; i < Type.arrayAnyValue.length; i++ ) {
           if ( (cti = til.get(Type.arrayAnyValue[i].getName())) == null )
              errorMsg.show("Class " + Type.arrayAnyValue[i].getName() + " not found");
           Type.arrayAnyValue[i].setTypeInterface(cti);
       }
       Type.anyTypeInterface = til.get("Any");
       Type.anyClassTypeInterface = til.get("AnyClass");
       Type.anyClassObjectTypeInterface = til.get("AnyClassObject");
       Type.anyValueTypeInterface = til.get("AnyValue");
       Type.reflectiveAnyClassTypeInterface = til.get("ReflectiveAnyClass");
       Type.nilTypeInterface = til.get("Nil");
       Type.catchUncheckedExceptionTypeInterface = til.get("CatchUncheckedException");
       Type.catchTypeInterface = til.get("Catch");
       Type.exceptionTypeInterface = til.get("Exception");
       /*
       Type.CharTypeInterface = til.get("Char");
       Type.BooleanTypeInterface = til.get("Boolean");
       Type.ByteTypeInterface = til.get("Byte");
       Type.IntegerTypeInterface  = til.get("Integer");
       Type.LongTypeInterface = til.get("Long");
       Type.RealTypeInterface = til.get("Real");
       Type.DoubleTypeInterface = til.get("Double");        
       */
       Type.StringTypeInterface = til.get("String");
       Type.wrapperCharTypeInterface = til.get("Char");
       Type.wrapperByteTypeInterface = til.get("Byte");
       Type.wrapperBooleanTypeInterface = til.get("Boolean");
       Type.wrapperIntegerTypeInterface = til.get("Integer");
       Type.wrapperLongTypeInterface = til.get("Long");
       Type.wrapperRealTypeInterface = til.get("Real");
       Type.wrapperDoubleTypeInterface = til.get("Double");
       
       
       if ( Type.anyTypeInterface == null || 
            Type.anyClassTypeInterface == null ||
            Type.anyClassObjectTypeInterface == null || 
            Type.anyValueTypeInterface == null ||
            Type.reflectiveAnyClassTypeInterface == null ||
            Type.nilTypeInterface == null ||
            /*
            Type.CharTypeInterface == null ||
            Type.BooleanTypeInterface == null ||
            Type.ByteTypeInterface == null || 
            Type.IntegerTypeInterface == null ||
            Type.LongTypeInterface == null ||
            Type.RealTypeInterface == null ||
            Type.DoubleTypeInterface == null ||
            */
            Type.StringTypeInterface == null ||
            Type.wrapperCharTypeInterface == null ||
            Type.wrapperByteTypeInterface == null ||
            Type.wrapperBooleanTypeInterface == null ||
            Type.wrapperIntegerTypeInterface == null ||
            Type.wrapperLongTypeInterface == null ||
            Type.wrapperRealTypeInterface == null ||
            Type.wrapperDoubleTypeInterface == null ||
            Type.catchUncheckedExceptionTypeInterface == null ||
            Type.catchTypeInterface == null ||
            Type.exceptionTypeInterface == null
            )
            
          errorMsg.show("One of the following classes was not found: Any, AnyClass, ... Char, Boolean, ... CatchUncheckedException");
          
        Type.nilType.setTypeInterface(Type.nilTypeInterface);
        
        Type.anyType.setTypeInterface(Type.anyTypeInterface);
        Type.anyClassType.setTypeInterface(Type.anyClassTypeInterface);
        Type.anyValueType.setTypeInterface(Type.anyValueTypeInterface);
        Type.anyClassObjectType.setTypeInterface(Type.anyClassObjectTypeInterface);
        Type.reflectiveAnyClassType.setTypeInterface(Type.reflectiveAnyClassTypeInterface);
        Type.wrapperCharType.setTypeInterface(Type.wrapperCharTypeInterface);
        Type.wrapperByteType.setTypeInterface(Type.wrapperByteTypeInterface);
        Type.wrapperBooleanType.setTypeInterface(Type.wrapperBooleanTypeInterface);
        Type.wrapperIntegerType.setTypeInterface(Type.wrapperIntegerTypeInterface);
        Type.wrapperLongType.setTypeInterface(Type.wrapperLongTypeInterface);
        Type.wrapperRealType.setTypeInterface(Type.wrapperRealTypeInterface);
        Type.wrapperDoubleType.setTypeInterface(Type.wrapperDoubleTypeInterface);
        Type.catchUncheckedExceptionType.setTypeInterface(Type.catchUncheckedExceptionTypeInterface);
        Type.catchType.setTypeInterface(Type.catchTypeInterface);
        Type.exceptionType.setTypeInterface(Type.exceptionTypeInterface);
    }
    
     
    private boolean addBasicLib() {
        
        if ( pathBasicLib == null ) {
            error.show("No path for basic library classes was specified", -1);
            return false;
        }
        else 
          return addClasses(pathBasicLib);
    }

    private boolean addWrapperLib() {
        
        if ( pathWrapper != null ) 
           return addClasses(pathWrapper);
        else
           return true;
    }

    private boolean addArrayLib() {

        boolean result = true;
        Enumeration e = arrayPathLib.elements();
        while ( e.hasMoreElements() ) 
           if ( ! addClasses( (String ) e.nextElement() ) ) 
               result = false;
        return result;
    }
    
    private boolean addClasses( String path ) {
        
        File file = new File(path);
        if ( ! file.exists() || ! file.canRead() || ! file.isDirectory() ) {
           error.show("The path " + path + 
              " does not exist or I can't read it or it is not a directory", -1);
           return false;
        }
        else {
           String allFiles[] = file.list();
           Hashtable fileNames = new Hashtable();
           for( int i = 0; i < allFiles.length; i++ ) 
              if ( (allFiles[i].endsWith(".g") || allFiles[i].endsWith(".gi")) &&
                   ! allFiles[i].startsWith("~") ) 
                  fileNames.put( 
                     allFiles[i].substring(0, allFiles[i].lastIndexOf('.')), "" );
           Enumeration e = fileNames.keys();
           while ( e.hasMoreElements() ) {
              String name = (String ) e.nextElement();
              cfl.addElement( new ClassFile( path, name ) );
           }
          return true;
       }
    }
    
    



    private boolean parse() {
        /*
          parse the text file of the project
        */
        try { 
           Vector list = fileNameList();
           if ( ! optionsList() ) return false;
           if ( list.size() < 1 )  {
              error.show("There must be at least one class in the program, besides the libraries");
              return false;
           }
           else {
              startClass = (String ) list.elementAt(0);
              if ( startClass.indexOf('.') >= 0 || 
                   startClass.indexOf(separator) >= 0 ) {
                 error.show("The first line in the project should contain a class name, not a file name");
                 return false;
              }
              Enumeration e = list.elements();
              while ( e.hasMoreElements() ) 
                 cfl.addElement( new ClassFile(projectDir, (String ) e.nextElement()) );
              if ( token.tk != PSymbol.EOF ) {
                 error.show("EOF expected", lineNumber);
                 return false;
              }
              else
                 if ( templateDir == null )  {
                    error.show("templateDir not specified in project");
                    return false;
                 }
           }
        } catch ( Exception excep ) { return false; } 
        return true;
    }
    
    private Vector fileNameList() {
        Vector list = new Vector();
        while ( token.tk == PSymbol.STRING || token.tk == PSymbol.PATHNAME ) {
            String name  = (String ) token.getValue();
            int i = name.lastIndexOf('.');
            if ( i >= 0 ) 
                name = name.substring(0, i);
            list.addElement(name);
            nextToken();
        }
        return list;
    }
            
    private boolean optionsList() {
        if ( token.tk != PSymbol.OPTIONS ) {
           error.show("Options expected", lineNumber);
           return false;
        }
        else {
            inFileName = false;
            nextToken();
            while ( token.tk != PSymbol.EOF )
              if ( ! option() ) 
                 return false;
            return true;
        }
    }

    
    private boolean option() {
        switch ( token.tk ) {
            case PSymbol.BASICLIB :
              nextToken();
              if ( token.tk != PSymbol.ASSIGN ) {
                error.show("= expected", lineNumber);
                return false;
              }
              else {
                inFileName = true;
                nextToken();
                inFileName = false;
                if ( token.tk != PSymbol.STRING && token.tk != PSymbol.PATHNAME ) {
                  error.show("File name expected", lineNumber);
                  return false;
                }
                else {
                    pathBasicLib = (String ) token.getValue();
                    nextToken();
                }
              }
              break;
            case PSymbol.LIB :
              nextToken();
              if ( token.tk != PSymbol.ASSIGN ) {
                error.show("= expected", lineNumber);
                return false;
              }
              else {
                inFileName = true;
                nextToken();
                inFileName = false;
                if ( token.tk != PSymbol.STRING && token.tk != PSymbol.PATHNAME ) {
                  error.show("File name expected", lineNumber);
                  return false;
                }
                else {
                    arrayPathLib.addElement( (String ) token.getValue() );
                    nextToken();
                }
              }
              break;
            case PSymbol.JAVALIB :
              nextToken();
              if ( token.tk != PSymbol.ASSIGN ) {
                error.show("= expected", lineNumber);
                return false;
              }
              else {
                inFileName = true;
                nextToken();
                inFileName = false;
                if ( token.tk != PSymbol.STRING && token.tk != PSymbol.PATHNAME ) {
                  error.show("File name expected", lineNumber);
                  return false;
                }
                else {
                    arrayJavaLib.addElement( (String ) token.getValue() );
                    nextToken();
                }
              }
              break;
            case PSymbol.WRAPPER:
              nextToken();
              if ( token.tk != PSymbol.ASSIGN ) {
                error.show("= expected", lineNumber);
                return false;
              }
              else {
                inFileName = true;
                nextToken();
                inFileName = false;
                if ( token.tk != PSymbol.STRING && token.tk != PSymbol.PATHNAME ) {
                  error.show("File name expected", lineNumber);
                  return false;
                }
                else {
                    pathWrapper = (String ) token.getValue();
                    nextToken();
                }
              }
              break;
            case PSymbol.TEMPLATEDIR : 
              nextToken();
              if ( token.tk != PSymbol.ASSIGN ) {
                error.show("= expected", lineNumber);
                return false;
              }
              else {
                inFileName = true;
                nextToken();
                inFileName = false;
                if ( token.tk != PSymbol.STRING && token.tk != PSymbol.PATHNAME ) {
                  error.show("File name expected", lineNumber);
                  return false;
                }
                else {
                    templateDir = (String ) token.getValue();
                    /* if ( templateDir.length > 0 && 
                         templateDir.charAt(templateDir.length - 1) != separator.charAt(0) )
                       templateDir = templateDir + separator; */
                    nextToken();
                }
              }
              break;
            case PSymbol.PROJECTDIR: 
              nextToken();
              if ( token.tk != PSymbol.ASSIGN ) {
                error.show("= expected", lineNumber);
                return false;
              }
              else {
                inFileName = true;
                nextToken();
                inFileName = false;
                if ( token.tk != PSymbol.STRING && token.tk != PSymbol.PATHNAME ) {
                  error.show("File name expected", lineNumber);
                  return false;
                }
                else {
                    projectDir = (String ) token.getValue();
                    nextToken();
                    Enumeration e = toBeCompiledList.elements();
                    while ( e.hasMoreElements() ) {
                       ClassFile cf = (ClassFile ) e.nextElement();
                       cf.setDirectory(projectDir);
                    }
                }
              }
              break;
            case PSymbol.USEIRL :
              compilerOptions.setUseIRL(true);
              nextToken();
              break;
            case PSymbol.COMPILEONLY :
              compilerOptions.setCompileOnly(true);
              nextToken();
              break;
            case PSymbol.COMPILEALL :
              compilerOptions.setCompileAll(true);
              nextToken();
              break;
            case PSymbol.ARRAYDIR :
              nextToken();
              if ( token.tk != PSymbol.ASSIGN ) {
                error.show("= expected", lineNumber);
                return false;
              }
              else {
                inFileName = true;
                nextToken();
                inFileName = false;
                if ( token.tk != PSymbol.STRING && token.tk != PSymbol.PATHNAME ) {
                  error.show("Directory name expected", lineNumber);
                  return false;
                }
                else {
                    arrayDir = (String ) token.getValue();
                    nextToken();
                }
              }
              break;
            case PSymbol.CLASSDIR:
              nextToken();
              if ( token.tk != PSymbol.ASSIGN ) {
                error.show("= expected", lineNumber);
                return false;
              }
              else {
                inFileName = true;
                nextToken();
                inFileName = false;
                if ( token.tk != PSymbol.STRING && token.tk != PSymbol.PATHNAME ) {
                  error.show("Directory name expected", lineNumber);
                  return false;
                }
                else {
                    classDir = (String ) token.getValue();
                    nextToken();
                }
              }
              break;
            case PSymbol.INTERFACEDIR :
              nextToken();
              if ( token.tk != PSymbol.ASSIGN ) {
                error.show("= expected", lineNumber);
                return false;
              }
              else {
                inFileName = true;
                nextToken();
                inFileName = false;
                if ( token.tk != PSymbol.STRING && token.tk != PSymbol.PATHNAME ) {
                  error.show("Directory name expected", lineNumber);
                  return false;
                }
                else {
                    interfaceDir = (String ) token.getValue();
                    nextToken();
                }
              }
              break;
            case PSymbol.SUBTYPEEQUALSUBCLASS : 
               subtypeEqualSubclass = true;
               break;
            default:
              error.show("Option expected", lineNumber);
              return false;
              
        }
        return true;
    }
              
        
        
        
    
    private void nextToken() {
        char ch;
        
        while (  (ch = input[tokenPos]) == ' ' || ch == '\r' ||
                 ch == '\t' || ch == '\n')  {
            // count the number of lines
          if ( ch == '\n')
            lineNumber++;
          tokenPos++;
          }
        if ( ch == '\0') 
          token = new PSymbol(PSymbol.EOF);
        else
          if ( input[tokenPos] == '/' && input[tokenPos + 1] == '/' ) {
                // comment found
               while ( input[tokenPos] != '\0'&& input[tokenPos] != '\n' )
                 tokenPos++;
               nextToken();
               }
          else {
            if ( inFileName ) {
                if ( ch == '\"' ) {
                    tokenPos++;
                    getString();
                }
                else {
                    StringBuffer ident = new StringBuffer();
                    while ( ! Character.isWhitespace(input[tokenPos]) 
                            && input[tokenPos] != '\0' ) {
                        ident.append( input[tokenPos] );
                        tokenPos++;
                    }
                    String s = ident.toString();
                    if ( beforeOptions && s.toLowerCase().compareTo("*options") == 0 ) {
                        beforeOptions = false;
                        token = new PSymbol( PSymbol.OPTIONS );
                    }
                    else
                        token = new PSymbol( PSymbol.PATHNAME, s );
                    }
            }
            else {

               if ( Character.isLetter( ch ) ) {
                   // get a keyword
                   StringBuffer ident = new StringBuffer();
                   while ( Character.isLetter( input[tokenPos] ) ) {
                       ident.append(input[tokenPos]);
                       tokenPos++;
                   }
                   String identStr = ident.toString().toLowerCase();
                      // if identStr is in the list of keywords, it is a keyword !
                   Object value = keywordsTable.get(identStr);
                   if ( value == null ) 
                     error.show( "Option " + ident + " is not valid", lineNumber );
                   else 
                       // get the token
                     token = (PSymbol ) value;
                   if ( Character.isDigit(input[tokenPos]) )
                     error.show("Word followed by a number", lineNumber);
               }
               else if ( Character.isDigit( ch ) ) {
                   // get a number
                   StringBuffer number = new StringBuffer();
                   while ( Character.isDigit( input[tokenPos] ) ) {
                       number.append(input[tokenPos]);
                       tokenPos++;
                   }
                   token =  new PSymbol( PSymbol.NUMBER, new Integer(number.toString()) );
               } else {
                   tokenPos++;
                   switch ( ch ) {
                       case '=' :
                         token = new PSymbol( PSymbol.ASSIGN );
                         break;
                       case '"' :
                          getString();
                          break;
                       default :
                         error.show("Invalid Character: '" + ch + "'", lineNumber);
                       }
                   }
               }
             }
    }

     
    private void getString() {
        
        StringBuffer ident = new StringBuffer();
        while ( input[tokenPos] != '\"' && input[tokenPos] != '\0' ) {
           ident.append(input[tokenPos]);
           tokenPos++;
        }
        String identStr = ident.toString();
        token = new PSymbol( PSymbol.STRING, identStr );
        if ( input[tokenPos] == '\0' ) 
           error.signal("End-of-file unexpected", lineNumber);
        else
           tokenPos++;
    }
    

      // contains the keywords
    static private Hashtable keywordsTable;

     // this code will be executed only once for each program execution
     static {
        keywordsTable = new Hashtable();
        keywordsTable.put( "options", new PSymbol(PSymbol.OPTIONS) );
        keywordsTable.put( "basiclib", new PSymbol(PSymbol.BASICLIB) );
        keywordsTable.put( "lib", new PSymbol(PSymbol.LIB) );
        keywordsTable.put( "wrapper", new PSymbol(PSymbol.WRAPPER) );
        keywordsTable.put( "useirl", new PSymbol(PSymbol.USEIRL) );
        keywordsTable.put( "compileonly", new PSymbol(PSymbol.COMPILEONLY) );
        keywordsTable.put( "compileall", new PSymbol(PSymbol.COMPILEALL) );
        keywordsTable.put( "arraydir", new PSymbol(PSymbol.ARRAYDIR) );
        keywordsTable.put( "classdir", new PSymbol(PSymbol.CLASSDIR) );
        keywordsTable.put( "interfacedir", new PSymbol(PSymbol.INTERFACEDIR) );
        keywordsTable.put( "templatedir", new PSymbol(PSymbol.TEMPLATEDIR) );
        keywordsTable.put( "projectdir", new PSymbol(PSymbol.PROJECTDIR) );
        keywordsTable.put( "subtypeequalsubclass", new PSymbol(PSymbol.SUBTYPEEQUALSUBCLASS) );
     }
     
    static private Hashtable basicClassesTable;
    static {
        basicClassesTable = new Hashtable();
        basicClassesTable.put("char", "");
        basicClassesTable.put("boolean", "");
        basicClassesTable.put("byte", "");
        basicClassesTable.put("integer", "");
        basicClassesTable.put("long", "");
        basicClassesTable.put("real", "");
        basicClassesTable.put("double", "");
        basicClassesTable.put("String", "");
    }
        
     
     
      // current token
    private PSymbol token;
    private int  tokenPos, lineNumber;
      // program given as input - source code
    private char []input;
    private String name, projectDir, separator;
    private ProjectError error;
    private ErrorMsgList errorMsg;
    private SymbolTable symbolTable;

    private String pathBasicLib; // basic classes and IRL
    private String pathWrapper;  // Wrapper classes
    
    private int state;
    
    private Vector arrayPathLib, arrayJavaLib, toBeCompiledList;
    //c private ClassList classList;
    private String arrayDir, classDir, templateDir, interfaceDir;
    char []contentsOf_array_any_1_g;
    char []contentsOf_iter_array_any_1_g;
    private ToCreateClassList toCreateClassList;
    private ToCreateArrayList toCreateArrayList;
    private boolean beforeOptions, inFileName;

    private Vector dirList;  // list of all directories that will contain java files 
      // classFileList. A list of all classes cited in the project
      // Those classes of cfl that have not been compiled will be
      // put at ccl.
    private Vector cfl; 
      // compileClassList. A list of classes to be compiled. Derived from cfl
    private Vector ccl; 
      // A list of parameterized classes. They have not been compiled
      // They will be used as templates to create classes like List(char)
    private Hashtable pcl; 
       // type interface list. A list of all interfaces of all 
       // classes used in the program.
    private TypeInterfaceList til; 
    
    private File fileAny1g; // one of the templates for arrays
    private File fileIterAny1g; // other !
    
    public static boolean subtypeEqualSubclass;
    
    private String startClass; // class in which the execution should begin
    ClassErrorListList errorListList; // list of lists of errors
    
    CompilerOptions compilerOptions;
}

    
